# -*- coding: utf-8 -*-
##################################################################################
#
#  © 2016 Farukh Tamboli. All Rights Reserved.
#
# NOTICE:  All information contained herein is, and remains the property of
# Farukh Tamboli. The intellectual and technical concepts contained herein are
# proprietary to Farukh Tamboli and may be covered by U.S. and Foreign Patents,
# patents in process, and are protected by trade secret or copyright law.
# Dissemination of this information or reproduction of this material is strictly
# forbidden unless prior written permission is obtained from Farukh Tamboli.
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
##################################################################################


import pymysql
from sqlalchemy import create_engine
import pandas as pd
import sys


def stdoutprint(str):
    try:
        print(str)
    except UnicodeEncodeError:
        print(str.encode(sys.stdout.encoding,errors='replace'))

def getConnectionSettings(env, db):
    key = env + '_' + db
    config = {
        'prod_db':
            {
                'host': "testdbserver",
                'user': "username",
                'password': "password",
                'db': "dbname"
            }
    }
    return config[key]

class Py3Cursor(pymysql.cursors.Cursor):
    def mogrify(self, query, args=None):
        """
        Returns the exact string that is sent to the database by calling the
        execute() method.

        This method follows the extension to the DB API 2.0 followed by Psycopg.
        """
        conn = self._get_db()
        #if PY2:  # Use bytes on Python 2 always
        #    query = self._ensure_bytes(query, encoding=conn.encoding)

        if args is not None:
            if query.find('%s')>0:
                query = query % self._escape_args(args, conn)
            else:
                args = self._escape_args(args, conn)
                if type(args) == dict:
                    query = query.format(**args)
                elif type(args) == tuple:
                    query = query.format(*args)
                else:
                    query = query.format(args)

        return query


class DatabaseConnection:

    def __init__(self, settings=None, **kwargs):
        if settings is None:
            settings = kwargs
        self.host = settings['host']
        self.user = settings['user']
        self.password = settings['password']
        self.db = settings['db']

        stdoutprint("DatabaseConnection: init\nServer: {}\nDatabase: {}\nUser: {}\n".format(self.host, self.db, self.user))
        self.conn = pymysql.connect(host=self.host, user=self.user, password=self.password, db=self.db, charset='utf8')

    def update(self, sqlStr, params=None):
        stdoutprint("DatabaseConnection: update\nSQL:\n{}\n".format(sqlStr))
        cursor = self.conn.cursor(cursor=Py3Cursor)
        cursor.execute(sqlStr, params)
        cursor.close()

    def deleteManyInList(self, sqlStr, id_list):
        id_list = "('" + "','".join(id_list) + "')"
        sqlStr = sqlStr.format(id_list)
        stdoutprint("DatabaseConnection: deleteManyInList\nSQL:\n{}\n".format(sqlStr))
        cursor = self.conn.cursor(cursor=Py3Cursor)
        cursor.execute(sqlStr)
        cursor.close()

    def insertGetId(self, sqlStr, params=None):
        stdoutprint("DatabaseConnection: insertGetId\nSQL:\n{}\n".format(sqlStr))
        cursor = self.conn.cursor(cursor=Py3Cursor)
        cursor.execute(sqlStr,params)
        newid = cursor.lastrowid
        cursor.close()
        return newid

    def selectScalar(self, sqlStr, params=None):
        cursor = self.conn.cursor(cursor=Py3Cursor)
        stdoutprint("DatabaseConnection: selectScalar\nSQL:\n{}\n".format(cursor.mogrify(sqlStr, params)))
        cursor.execute(sqlStr, params)
        row = cursor.fetchone()
        if row is None:
            newret = None
        else:
            newret = row[0]
        cursor.close()
        return newret

    def selectScalarList(self, sqlStr, params=None):
        stdoutprint("DatabaseConnection: selectScalarList\nSQL:\n{}\n".format(sqlStr))
        cursor = self.conn.cursor(cursor=Py3Cursor)
        cursor.execute(sqlStr, params)
        newret = []
        for row in cursor.fetchall():
            newret.append(row[0])
        cursor.close()
        return newret

    def getColumns(self, sqlStr, params=None):
        stdoutprint("DatabaseConnection: getColumns\nSQL:\n{}\n".format(sqlStr,params))
        cursor = self.conn.cursor(cursor=Py3Cursor, buffered=True)
        cursor.execute(sqlStr, params)
        field_names = [i[0] for i in cursor.description]
        cursor.close()
        return field_names

    def selectOneRow(self, sqlStr, params=None):
        stdoutprint("DatabaseConnection: selectOneRow\nSQL:\n{}\n".format(sqlStr))
        cursor = self.conn.cursor(cursor=Py3Cursor)
        cursor.execute(sqlStr, params)
        num_fields = len(cursor.description)
        field_names = [i[0] for i in cursor.description]
        newret = {}
        row = cursor.fetchone()
        for i in range(num_fields):
            newret[field_names[i]] = row[i]
        cursor.close()
        return newret

    def selectManyRows(self, sqlStr, params=None):
        stdoutprint("DatabaseConnection: selectManyRows\nSQL:\n{}\n".format(sqlStr))
        cursor = self.conn.cursor(cursor=Py3Cursor)
        if params:
            #sqlStr = format(sqlStr,params)
            sqlStr = cursor.mogrify(sqlStr, params)

        print(params)
        print(sqlStr)

        cursor.execute(sqlStr)
        num_fields = len(cursor.description)
        field_names = [i[0] for i in cursor.description]
        newret = []
        for row in cursor.fetchall():
            newrow = {}
            for i in range(num_fields):
                newrow[field_names[i]] = row[i]
            newret.append(newrow)
        cursor.close()
        return newret

    def getDataFrame(self, sqlStr, params=None, *args, **kwargs):
        cursor = Py3Cursor(self.conn)
        query = cursor.mogrify(sqlStr, params)
        stdoutprint(query)
        return pd.read_sql(query, con=self.conn, params=None, coerce_float=True, *args, **kwargs)

    # def getDataFrame(self, sqlStr, params=None, chunksize=None):
    #     cursor = Py3Cursor(self.conn)
    #     query = cursor.mogrify(sqlStr, params)
    #     stdoutprint(query)
    #     chunks = pd.read_sql(query, con=self.conn, params=None, coerce_float=True, chunksize=chunksize)
    #     if chunksize is None:
    #         return chunks
    #     else:
    #         out = []
    #         for chunk in chunks:
    #             out.append(chunk)
    #         return pd.concat(out)

    def writeDataFrame(self, data, table, method='append', chunksize=None, index=False, index_label=None, dtype=None):
        # writeDataFrame uses SQLAlchemy to connect to the database and not PyMySQL
        engine = create_engine("mysql+pymysql://" + self.user + ":" + self.password + "@" + self.host + "/" + self.db)
        stdoutprint("DatabaseConnection: writeDataFrame\nTable: {}\nData Frame:\n{}\n".format(table, data))
        if method not in ("replace", "append"):
            method = "fail"
        data.to_sql(name=table, con=engine, if_exists=method, index=index, chunksize=chunksize,
                    index_label=index_label, dtype=dtype)

    def commit(self):
        print("DatabaseConnection: commit\n")
        self.conn.commit()

    def rollback(self):
        print("DatabaseConnection: rollback\n")
        self.conn.rollback()

    def destroy(self):
        if self.conn is not None:
            stdoutprint("DatabaseConnection: destroy\nServer: {}\nDatabase: {}\nUser: {}\n"
                  .format(self.host, self.db, self.user))
            self.conn.close()
            self.conn = None

    def __del__(self):
        self.destroy()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.destroy()


import redis
import pickle


class KeyStoreServer:
    def __init__(self):
        self.redisConn = redis.StrictRedis(host='localhost', port=6379, db=0)

    def getDataFrame(self, key):
        return pd.read_msgpack(self.redisConn.get(key))

    def setDataFrame(self, key, df):
        self.redisConn.set(key, df.to_msgpack(compress='zlib'))

    def getObject(self, key):
        return pickle.loads(self.redisConn.get(key))

    def setObject(self, key, obj):
        self.redisConn.set(key, pickle.dumps(obj))

    def setDataFrameIfNotExists(self, key, df):
        self.redisConn.setnx(key, df.to_msgpack(compress='zlib'))

    def set(self,key, value):
        self.redisConn.set(key, value)

    def setIfNotExists(self,key, value):
        self.redisConn.setnx(key, value)

    def get(self, key):
        return self.redisConn.get(key)

    def isKeyExists(self, key):
        return self.redisConn.exists(key)

    def getPipeline(self, transaction=True, shard_hint=None):
        return self.redisConn.pipeline(transaction, shard_hint)

    def keyFromObject(self, obj):
        return pickle.dumps(obj)

    def objectFromKey(self, key):
        return pickle.loads(key)

    def __getattr__(self, key):
        if key in self:
            value = self.redisConn.get(key)
            if isinstance(value, pd.DataFrame):
                return pd.read_msgpack(value)
            elif isinstance(value, object):
                return pickle.loads(value)
            else:
                return value
        else:
            raise AttributeError("No such name: " + key)

    def __setattr__(self, key, value):
        if isinstance(value, pd.DataFrame):
            self.redisConn.set(key, value.to_msgpack(compress='zlib'))
        elif isinstance(value, object):
            self.redisConn.set(key, pickle.dumps(value))
        else:
            self.redisConn.set(key, value)
